@@ -1,6 +1,6 @@
 /*
  *  R : A Computer Language for Statistical Data Analysis
- *  Copyright (C) 1999-2013  The R Core Team
+ *  Copyright (C) 1999-2014  The R Core Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -14,7 +14,7 @@
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, a copy is available at
- *  http://www.r-project.org/Licenses/
+ *  https://www.R-project.org/Licenses/
  */
 
 #ifdef HAVE_CONFIG_H
@@ -279,7 +279,7 @@ void nmmin(int n, double *Bvec, double *X, double *Fmin, optimfn fminfn,
     double oldsize;
     double **P;
     double size, step, temp, trystep;
-    char tstr[9]; // allow for 10^8 iters ...
+    char tstr[12]; // allow for 10^8 iters and pacify gcc7
     double VH, VL, VR;
 
     if (maxit <= 0) {
@@ -368,7 +368,7 @@ void nmmin(int n, double *Bvec, double *X, double *Fmin, optimfn fminfn,
 
 	    // avoid buffer overflow at 100001 iters. (PR#15240)
 	    if (trace) {
-		snprintf(tstr, 9, "%5d", funcount);
+		snprintf(tstr, 12, "%5d", funcount);
 		Rprintf("%s%s %f %f\n", action, tstr, VH, VL);
 	    }
 
