@@ -1,7 +1,7 @@
 /*
  *  R : A Computer Language for Statistical Data Analysis
+ *  Copyright (C) 1997--2021  The R Core Team
  *  Copyright (C) 1995, 1996  Robert Gentleman and Ross Ihaka
- *  Copyright (C) 1997--2013  The R Core Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -15,7 +15,7 @@
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, a copy is available at
- *  http://www.r-project.org/Licenses/
+ *  https://www.R-project.org/Licenses/
  */
 
 
@@ -27,32 +27,31 @@
 #include <Internal.h>
 #include <Rmath.h>
 #include <errno.h>
+#include <R_ext/Itermacros.h>
 
 /* interval at which to check interrupts, a guess */
 #define NINTERRUPT 10000000
 
-static SEXP integer_relop(RELOP_TYPE code, SEXP s1, SEXP s2);
-static SEXP real_relop(RELOP_TYPE code, SEXP s1, SEXP s2);
+static SEXP numeric_relop(RELOP_TYPE code, SEXP s1, SEXP s2);
 static SEXP complex_relop(RELOP_TYPE code, SEXP s1, SEXP s2, SEXP call);
-static SEXP string_relop(RELOP_TYPE code, SEXP s1, SEXP s2);
-static SEXP raw_relop(RELOP_TYPE code, SEXP s1, SEXP s2);
+static SEXP string_relop (RELOP_TYPE code, SEXP s1, SEXP s2);
+static SEXP raw_relop    (RELOP_TYPE code, SEXP s1, SEXP s2);
 
-#define DO_SCALAR_RELOP(oper, x, y) do {			\
-	switch (oper) {						\
-	case EQOP: return ScalarLogical((x) == (y));		\
-	case NEOP: return ScalarLogical((x) != (y));		\
-	case LTOP: return ScalarLogical((x) < (y));		\
-	case GTOP: return ScalarLogical((x) > (y));		\
-	case LEOP: return ScalarLogical((x) <= (y));		\
-	case GEOP: return ScalarLogical((x) >= (y));		\
-	}							\
+#define DO_SCALAR_RELOP(oper, x, y) do {		\
+	switch (oper) {					\
+	case EQOP: return ScalarLogical((x) == (y));	\
+	case NEOP: return ScalarLogical((x) != (y));	\
+	case LTOP: return ScalarLogical((x) < (y));	\
+	case GTOP: return ScalarLogical((x) > (y));	\
+	case LEOP: return ScalarLogical((x) <= (y));	\
+	case GEOP: return ScalarLogical((x) >= (y));	\
+	}						\
     } while (0)
 
 SEXP attribute_hidden do_relop(SEXP call, SEXP op, SEXP args, SEXP env)
 {
     SEXP ans, arg1, arg2;
     int argc;
-    int oper = PRIMVAL(op);
 
     if (args != R_NilValue &&
 	CDR(args) != R_NilValue &&
@@ -67,113 +66,75 @@ SEXP attribute_hidden do_relop(SEXP call, SEXP op, SEXP args, SEXP env)
 	if (DispatchGroup("Ops", call, op, args, env, &ans))
 	    return ans;
     }
-    else if (argc == 2) {
-	if (IS_SCALAR(arg1, INTSXP)) {
-	    int ix = INTEGER(arg1)[0];
-	    if (IS_SCALAR(arg2, INTSXP)) {
-		int iy = INTEGER(arg2)[0];
-		if (ix == NA_INTEGER || iy == NA_INTEGER)
-		    return ScalarLogical(NA_LOGICAL);
-		DO_SCALAR_RELOP(oper, ix, iy);
-	    }
-	    else if (IS_SCALAR(arg2, REALSXP)) {
-		double dy = REAL(arg2)[0];
-		if (ix == NA_INTEGER || ISNAN(dy))
-		    return ScalarLogical(NA_LOGICAL);
-		DO_SCALAR_RELOP(oper, ix, dy);
-	    }
-	}
-	else if (IS_SCALAR(arg1, REALSXP)) {
-	    double dx = REAL(arg1)[0];
-	    if (IS_SCALAR(arg2, INTSXP)) {
-		int iy = INTEGER(arg2)[0];
-		if (ISNAN(dx) || iy == NA_INTEGER)
-		    return ScalarLogical(NA_LOGICAL);
-		DO_SCALAR_RELOP(oper, dx, iy);
-	    }
-	    else if (IS_SCALAR(arg2, REALSXP)) {
-		double dy = REAL(arg2)[0];
-		if (ISNAN(dx) || ISNAN(dy))
-		    return ScalarLogical(NA_LOGICAL);
-		DO_SCALAR_RELOP(oper, dx, dy);
-	    }
-	}
-    }
-    else error("operator needs two arguments");
+
+    if (argc != 2)
+	error("operator needs two arguments");
 
     return do_relop_dflt(call, op, arg1, arg2);
 }
 
+// also called from cmp_relop() in eval.c :
 SEXP attribute_hidden do_relop_dflt(SEXP call, SEXP op, SEXP x, SEXP y)
 {
-    SEXP klass = R_NilValue, dims, tsp = R_NilValue, xnames, ynames;
-    R_xlen_t nx, ny;
-    int xarray, yarray, xts, yts;
-    Rboolean mismatch = FALSE, iS;
-    PROTECT_INDEX xpi, ypi;
-    SEXPTYPE typex, typey;
+    /* handle the REALSXP/INTSXP simple scalar case quickly */
+    if (IS_SIMPLE_SCALAR(x, INTSXP)) {
+	int ix = SCALAR_IVAL(x);
+	if (IS_SIMPLE_SCALAR(y, INTSXP)) {
+	    int iy = SCALAR_IVAL(y);
+	    if (ix == NA_INTEGER || iy == NA_INTEGER)
+		return ScalarLogical(NA_LOGICAL);
+	    DO_SCALAR_RELOP(PRIMVAL(op), ix, iy);
+	}
+	else if (IS_SIMPLE_SCALAR(y, REALSXP)) {
+	    double dy = SCALAR_DVAL(y);
+	    if (ix == NA_INTEGER || ISNAN(dy))
+		return ScalarLogical(NA_LOGICAL);
+	    DO_SCALAR_RELOP(PRIMVAL(op), ix, dy);
+	}
+    }
+    else if (IS_SIMPLE_SCALAR(x, REALSXP)) {
+	double dx = SCALAR_DVAL(x);
+	if (IS_SIMPLE_SCALAR(y, INTSXP)) {
+	    int iy = SCALAR_IVAL(y);
+	    if (ISNAN(dx) || iy == NA_INTEGER)
+		return ScalarLogical(NA_LOGICAL);
+	    DO_SCALAR_RELOP(PRIMVAL(op), dx, iy);
+	}
+	else if (IS_SIMPLE_SCALAR(y, REALSXP)) {
+	    double dy = SCALAR_DVAL(y);
+	    if (ISNAN(dx) || ISNAN(dy))
+		return ScalarLogical(NA_LOGICAL);
+	    DO_SCALAR_RELOP(PRIMVAL(op), dx, dy);
+	}
+    }
 
-    nx = xlength(x);
-    ny = xlength(y);
-    typex = TYPEOF(x);
-    typey = TYPEOF(y);
+    R_xlen_t
+	nx = xlength(x),
+	ny = xlength(y);
+    SEXPTYPE
+	typex = TYPEOF(x),
+	typey = TYPEOF(y);
 
-    /* pre-test to handle the most common case quickly.
-       Used to skip warning too ....
-     */
+    /* handle the REALSXP/INTSXP simple vector/scalar case quickly. */
     if (ATTRIB(x) == R_NilValue && ATTRIB(y) == R_NilValue &&
 	(typex == REALSXP || typex == INTSXP) &&
 	(typey == REALSXP || typey == INTSXP) &&
-	nx > 0 && ny > 0) {
+	nx > 0 && ny > 0 && (nx == 1 || ny == 1)) {
 
-	/* handle the scalar case */
-	if (nx == 1 && ny == 1) {
-	    if (typex == INTSXP && typey == INTSXP) {
-		int ix = INTEGER(x)[0];
-		int iy = INTEGER(y)[0];
-		if (ix == NA_INTEGER || iy == NA_INTEGER)
-		    return ScalarLogical(NA_LOGICAL);
-		DO_SCALAR_RELOP(PRIMVAL(op), ix, iy);
-	    }
-	    else {
-		double dx = typex == REALSXP ? REAL(x)[0] :
-		    INTEGER(x)[0] != NA_INTEGER ? INTEGER(x)[0] : NA_REAL;
-		double dy = typey == REALSXP ? REAL(y)[0] :
-		    INTEGER(y)[0] != NA_INTEGER ? INTEGER(y)[0] : NA_REAL;
-		if (ISNAN(dx) || ISNAN(dy))
-		    return ScalarLogical(NA_LOGICAL);
-		DO_SCALAR_RELOP(PRIMVAL(op), dx, dy);
-	    }
-	}
-
-	/* non-scalar case */
-	PROTECT_WITH_INDEX(x, &xpi);
-	PROTECT_WITH_INDEX(y, &ypi);
+	PROTECT(x);
+	PROTECT(y);
 	SEXP ans;
-	if (typex == INTSXP && typey == INTSXP) 
-	    ans = integer_relop(PRIMVAL(op), x, y);
-	else {
-	    if (typex == INTSXP)
-		REPROTECT(x = coerceVector(x, REALSXP), xpi);
-	    if (typey == INTSXP)
-		REPROTECT(y = coerceVector(y, REALSXP), ypi);
-	    ans = real_relop(PRIMVAL(op), x, y);
-	}
-	if (nx > 0 && ny > 0)
-	    mismatch = ((nx > ny) ? nx % ny : ny % nx) != 0;
-	if (mismatch) {
-	    PROTECT(ans);
-	    warningcall(call, _("longer object length is not "
-				"a multiple of shorter object length"));
-	    UNPROTECT(1);
-	}
+	ans = numeric_relop(PRIMVAL(op), x, y);
 	UNPROTECT(2);
 	return ans;
     }
 
+    /* handle the general case */
+    PROTECT_INDEX xpi, ypi;
     PROTECT_WITH_INDEX(x, &xpi);
     PROTECT_WITH_INDEX(y, &ypi);
 
+    Rboolean iS;
     /* That symbols and calls were allowed was undocumented prior to
        R 2.5.0.  We deparse them as deparse() would, minus attributes */
     if ((iS = isSymbol(x)) || TYPEOF(x) == LANGSXP) {
@@ -182,6 +143,7 @@ SEXP attribute_hidden do_relop_dflt(SEXP call, SEXP op, SEXP x, SEXP y)
 	SET_STRING_ELT(tmp, 0, (iS) ? PRINTNAME(x) :
 		       STRING_ELT(deparse1(x, 0, DEFAULTDEPARSE), 0));
 	REPROTECT(x = tmp, xpi);
+	nx = xlength(x);
 	UNPROTECT(1);
     }
     if ((iS = isSymbol(y)) || TYPEOF(y) == LANGSXP) {
@@ -190,49 +152,42 @@ SEXP attribute_hidden do_relop_dflt(SEXP call, SEXP op, SEXP x, SEXP y)
 	SET_STRING_ELT(tmp, 0, (iS) ? PRINTNAME(y) :
 		       STRING_ELT(deparse1(y, 0, DEFAULTDEPARSE), 0));
 	REPROTECT(y = tmp, ypi);
+	ny = xlength(y);
 	UNPROTECT(1);
     }
 
-    if (!isVector(x) || !isVector(y)) {
-	if (isNull(x) || isNull(y)) {
-	    UNPROTECT(2);
-	    return allocVector(LGLSXP,0);
-	}
+    if (isNull(x)) REPROTECT(x = allocVector(INTSXP,0), xpi);
+    if (isNull(y)) REPROTECT(y = allocVector(INTSXP,0), ypi);
+    if (!isVector(x) || !isVector(y))
 	errorcall(call,
 		  _("comparison (%d) is possible only for atomic and list types"),
 		  PRIMVAL(op));
-    }
 
     if (TYPEOF(x) == EXPRSXP || TYPEOF(y) == EXPRSXP)
 	errorcall(call, _("comparison is not allowed for expressions"));
 
     /* ELSE :  x and y are both atomic or list */
 
-    if (XLENGTH(x) <= 0 || XLENGTH(y) <= 0) {
-	UNPROTECT(2);
-	return allocVector(LGLSXP, 0);
-    }
-
-    mismatch = FALSE;
-    xarray = isArray(x);
-    yarray = isArray(y);
-    xts = isTs(x);
-    yts = isTs(y);
-    if (nx > 0 && ny > 0)
-	mismatch = ((nx > ny) ? nx % ny : ny % nx) != 0;
-
+    Rboolean
+	xarray = isArray(x),
+	yarray = isArray(y),
+	xts = isTs(x),
+	yts = isTs(y);
+    SEXP dims, xnames, ynames;
     if (xarray || yarray) {
 	if (xarray && yarray) {
 	    if (!conformable(x, y))
 		errorcall(call, _("non-conformable arrays"));
 	    PROTECT(dims = getAttrib(x, R_DimSymbol));
 	}
-	else if (xarray) {
+	else if (xarray && (ny != 0 || nx == 0)) {
 	    PROTECT(dims = getAttrib(x, R_DimSymbol));
 	}
-	else /*(yarray)*/ {
+	else if (yarray && (nx != 0 || ny == 0)) {
 	    PROTECT(dims = getAttrib(y, R_DimSymbol));
-	}
+	} else
+	    PROTECT(dims = R_NilValue);
+
 	PROTECT(xnames = getAttrib(x, R_DimNamesSymbol));
 	PROTECT(ynames = getAttrib(y, R_DimNamesSymbol));
     }
@@ -241,10 +196,11 @@ SEXP attribute_hidden do_relop_dflt(SEXP call, SEXP op, SEXP x, SEXP y)
 	PROTECT(xnames = getAttrib(x, R_NamesSymbol));
 	PROTECT(ynames = getAttrib(y, R_NamesSymbol));
     }
+
+    SEXP klass = NULL, tsp = NULL; // -Wall
     if (xts || yts) {
 	if (xts && yts) {
-	    if (!tsConform(x, y))
-		errorcall(call, _("non-conformable time series"));
+	    /* could check ts conformance here */
 	    PROTECT(tsp = getAttrib(x, R_TspSymbol));
 	    PROTECT(klass = getAttrib(x, R_ClassSymbol));
 	}
@@ -261,8 +217,11 @@ SEXP attribute_hidden do_relop_dflt(SEXP call, SEXP op, SEXP x, SEXP y)
 	    PROTECT(klass = getAttrib(y, R_ClassSymbol));
 	}
     }
-    if (mismatch)
-	warningcall(call, _("longer object length is not a multiple of shorter object length"));
+
+  if (nx > 0 && ny > 0) {
+	if(((nx > ny) ? nx % ny : ny % nx) != 0) // mismatch
+            warningcall(call, _(
+		"longer object length is not a multiple of shorter object length"));
 
     if (isString(x) || isString(y)) {
 	REPROTECT(x = coerceVector(x, STRSXP), xpi);
@@ -274,27 +233,32 @@ SEXP attribute_hidden do_relop_dflt(SEXP call, SEXP op, SEXP x, SEXP y)
 	REPROTECT(y = coerceVector(y, CPLXSXP), ypi);
 	x = complex_relop((RELOP_TYPE) PRIMVAL(op), x, y, call);
     }
+    else if ((isNumeric(x) || isLogical(x)) && (isNumeric(y) || isLogical(y))) {
+        x = numeric_relop((RELOP_TYPE) PRIMVAL(op), x, y);
+    } // rest of cases only apply when 'x' or 'y' is raw
     else if (isReal(x) || isReal(y)) {
 	REPROTECT(x = coerceVector(x, REALSXP), xpi);
 	REPROTECT(y = coerceVector(y, REALSXP), ypi);
-	x = real_relop((RELOP_TYPE) PRIMVAL(op), x, y);
+	x = numeric_relop((RELOP_TYPE) PRIMVAL(op), x, y);
     }
     else if (isInteger(x) || isInteger(y)) {
 	REPROTECT(x = coerceVector(x, INTSXP), xpi);
 	REPROTECT(y = coerceVector(y, INTSXP), ypi);
-	x = integer_relop((RELOP_TYPE) PRIMVAL(op), x, y);
+	x = numeric_relop((RELOP_TYPE) PRIMVAL(op), x, y);
     }
     else if (isLogical(x) || isLogical(y)) {
 	REPROTECT(x = coerceVector(x, LGLSXP), xpi);
 	REPROTECT(y = coerceVector(y, LGLSXP), ypi);
-	x = integer_relop((RELOP_TYPE) PRIMVAL(op), x, y);
+	x = numeric_relop((RELOP_TYPE) PRIMVAL(op), x, y);
     }
     else if (TYPEOF(x) == RAWSXP || TYPEOF(y) == RAWSXP) {
 	REPROTECT(x = coerceVector(x, RAWSXP), xpi);
 	REPROTECT(y = coerceVector(y, RAWSXP), ypi);
 	x = raw_relop((RELOP_TYPE) PRIMVAL(op), x, y);
     } else errorcall(call, _("comparison of these types is not implemented"));
-
+  } else { // nx == 0 || ny == 0
+	x = allocVector(LGLSXP, 0);
+  }
 
     PROTECT(x);
     if (dims != R_NilValue) {
@@ -305,9 +269,9 @@ SEXP attribute_hidden do_relop_dflt(SEXP call, SEXP op, SEXP x, SEXP y)
 	    setAttrib(x, R_DimNamesSymbol, ynames);
     }
     else {
-	if (xlength(x) == xlength(xnames))
+	if (xnames != R_NilValue && xlength(x) == xlength(xnames))
 	    setAttrib(x, R_NamesSymbol, xnames);
-	else if (xlength(x) == xlength(ynames))
+	else if (ynames != R_NilValue && xlength(x) == xlength(ynames))
 	    setAttrib(x, R_NamesSymbol, ynames);
     }
     if (xts || yts) {
@@ -320,19 +284,48 @@ SEXP attribute_hidden do_relop_dflt(SEXP call, SEXP op, SEXP x, SEXP y)
     return x;
 }
 
-/* i1 = i % n1; i2 = i % n2;
- * this macro is quite a bit faster than having real modulo calls
- * in the loop (tested on Intel and Sparc)
- */
-#define mod_iterate(n1,n2,i1,i2) for (i=i1=i2=0; i<n; \
-	i1 = (++i1 == n1) ? 0 : i1,\
-	i2 = (++i2 == n2) ? 0 : i2,\
-	++i)
+#define ISNA_INT(x) x == NA_INTEGER
 
-static SEXP integer_relop(RELOP_TYPE code, SEXP s1, SEXP s2)
+#define NR_HELPER(OP, type1, ACCESSOR1, ISNA1, type2, ACCESSOR2, ISNA2) do { \
+	type1 x1, *px1 = ACCESSOR1(s1);					\
+	type2 x2, *px2 = ACCESSOR2(s2);					\
+	int *pa = LOGICAL(ans);						\
+        MOD_ITERATE2(n, n1, n2, i, i1, i2, {                            \
+	    x1 = px1[i1];						\
+	    x2 = px2[i2];						\
+            if (ISNA1(x1) || ISNA2(x2))                                 \
+                pa[i] = NA_LOGICAL;					\
+            else                                                        \
+                pa[i] = (x1 OP x2);					\
+        });                                                             \
+    } while (0)
+
+#define NUMERIC_RELOP(type1, ACCESSOR1, ISNA1, type2, ACCESSOR2, ISNA2) do { \
+    switch (code) {                                                     \
+    case EQOP:                                                          \
+	NR_HELPER(==, type1, ACCESSOR1, ISNA1, type2, ACCESSOR2, ISNA2); \
+        break;                                                          \
+    case NEOP:                                                          \
+	NR_HELPER(!=, type1, ACCESSOR1, ISNA1, type2, ACCESSOR2, ISNA2); \
+        break;                                                          \
+    case LTOP:                                                          \
+	NR_HELPER(<, type1, ACCESSOR1, ISNA1, type2, ACCESSOR2, ISNA2); \
+        break;                                                          \
+    case GTOP:                                                          \
+	NR_HELPER(>, type1, ACCESSOR1, ISNA1, type2, ACCESSOR2, ISNA2); \
+        break;                                                          \
+    case LEOP:                                                          \
+	NR_HELPER(<=, type1, ACCESSOR1, ISNA1, type2, ACCESSOR2, ISNA2); \
+        break;                                                          \
+    case GEOP:                                                          \
+	NR_HELPER(>=, type1, ACCESSOR1, ISNA1, type2, ACCESSOR2, ISNA2); \
+        break;                                                          \
+    }                                                                   \
+} while(0)
+
+static SEXP numeric_relop(RELOP_TYPE code, SEXP s1, SEXP s2)
 {
     R_xlen_t i, i1, i2, n, n1, n2;
-    int x1, x2;
     SEXP ans;
 
     n1 = XLENGTH(s1);
@@ -342,159 +335,18 @@ static SEXP integer_relop(RELOP_TYPE code, SEXP s1, SEXP s2)
     PROTECT(s2);
     ans = allocVector(LGLSXP, n);
 
-    switch (code) {
-    case EQOP:
-	mod_iterate(n1, n2, i1, i2) {
-//	    if ((i+1) % NINTERRUPT == 0) R_CheckUserInterrupt();
-	    x1 = INTEGER(s1)[i1];
-	    x2 = INTEGER(s2)[i2];
-	    if (x1 == NA_INTEGER || x2 == NA_INTEGER)
-		LOGICAL(ans)[i] = NA_LOGICAL;
-	    else
-		LOGICAL(ans)[i] = (x1 == x2);
-	}
-	break;
-    case NEOP:
-	mod_iterate(n1, n2, i1, i2) {
-//	    if ((i+1) % NINTERRUPT == 0) R_CheckUserInterrupt();
-	    x1 = INTEGER(s1)[i1];
-	    x2 = INTEGER(s2)[i2];
-	    if (x1 == NA_INTEGER || x2 == NA_INTEGER)
-		LOGICAL(ans)[i] = NA_LOGICAL;
-	    else
-		LOGICAL(ans)[i] = (x1 != x2);
-	}
-	break;
-    case LTOP:
-	mod_iterate(n1, n2, i1, i2) {
-//	    if ((i+1) % NINTERRUPT == 0) R_CheckUserInterrupt();
-	    x1 = INTEGER(s1)[i1];
-	    x2 = INTEGER(s2)[i2];
-	    if (x1 == NA_INTEGER || x2 == NA_INTEGER)
-		LOGICAL(ans)[i] = NA_LOGICAL;
-	    else
-		LOGICAL(ans)[i] = (x1 < x2);
-	}
-	break;
-    case GTOP:
-	mod_iterate(n1, n2, i1, i2) {
-//	    if ((i+1) % NINTERRUPT == 0) R_CheckUserInterrupt();
-	    x1 = INTEGER(s1)[i1];
-	    x2 = INTEGER(s2)[i2];
-	    if (x1 == NA_INTEGER || x2 == NA_INTEGER)
-		LOGICAL(ans)[i] = NA_LOGICAL;
-	    else
-		LOGICAL(ans)[i] = (x1 > x2);
-	}
-	break;
-    case LEOP:
-	mod_iterate(n1, n2, i1, i2) {
-//	    if ((i+1) % NINTERRUPT == 0) R_CheckUserInterrupt();
-	    x1 = INTEGER(s1)[i1];
-	    x2 = INTEGER(s2)[i2];
-	    if (x1 == NA_INTEGER || x2 == NA_INTEGER)
-		LOGICAL(ans)[i] = NA_LOGICAL;
-	    else
-		LOGICAL(ans)[i] = (x1 <= x2);
-	}
-	break;
-    case GEOP:
-	mod_iterate(n1, n2, i1, i2) {
-//	    if ((i+1) % NINTERRUPT == 0) R_CheckUserInterrupt();
-	    x1 = INTEGER(s1)[i1];
-	    x2 = INTEGER(s2)[i2];
-	    if (x1 == NA_INTEGER || x2 == NA_INTEGER)
-		LOGICAL(ans)[i] = NA_LOGICAL;
-	    else
-		LOGICAL(ans)[i] = (x1 >= x2);
-	}
-	break;
+    if (isInteger(s1) || isLogical(s1)) {
+        if (isInteger(s2) || isLogical(s2)) {
+            NUMERIC_RELOP(int, INTEGER, ISNA_INT, int, INTEGER, ISNA_INT);
+        } else {
+            NUMERIC_RELOP(int, INTEGER, ISNA_INT, double, REAL, ISNAN);
+        }
+    } else if (isInteger(s2) || isLogical(s2)) {
+        NUMERIC_RELOP(double, REAL, ISNAN, int, INTEGER, ISNA_INT);
+    } else {
+        NUMERIC_RELOP(double, REAL, ISNAN, double, REAL, ISNAN);
     }
-    UNPROTECT(2);
-    return ans;
-}
 
-static SEXP real_relop(RELOP_TYPE code, SEXP s1, SEXP s2)
-{
-    R_xlen_t i, i1, i2, n, n1, n2;
-    double x1, x2;
-    SEXP ans;
-
-    n1 = XLENGTH(s1);
-    n2 = XLENGTH(s2);
-    n = (n1 > n2) ? n1 : n2;
-    PROTECT(s1);
-    PROTECT(s2);
-    ans = allocVector(LGLSXP, n);
-
-    switch (code) {
-    case EQOP:
-	mod_iterate(n1, n2, i1, i2) {
-//	    if ((i+1) % NINTERRUPT == 0) R_CheckUserInterrupt();
-	    x1 = REAL(s1)[i1];
-	    x2 = REAL(s2)[i2];
-	    if (ISNAN(x1) || ISNAN(x2))
-		LOGICAL(ans)[i] = NA_LOGICAL;
-	    else
-		LOGICAL(ans)[i] = (x1 == x2);
-	}
-	break;
-    case NEOP:
-	mod_iterate(n1, n2, i1, i2) {
-//	    if ((i+1) % NINTERRUPT == 0) R_CheckUserInterrupt();
-	    x1 = REAL(s1)[i1];
-	    x2 = REAL(s2)[i2];
-	    if (ISNAN(x1) || ISNAN(x2))
-		LOGICAL(ans)[i] = NA_LOGICAL;
-	    else
-		LOGICAL(ans)[i] = (x1 != x2);
-	}
-	break;
-    case LTOP:
-	mod_iterate(n1, n2, i1, i2) {
-//	    if ((i+1) % NINTERRUPT == 0) R_CheckUserInterrupt();
-	    x1 = REAL(s1)[i1];
-	    x2 = REAL(s2)[i2];
-	    if (ISNAN(x1) || ISNAN(x2))
-		LOGICAL(ans)[i] = NA_LOGICAL;
-	    else
-		LOGICAL(ans)[i] = (x1 < x2);
-	}
-	break;
-    case GTOP:
-	mod_iterate(n1, n2, i1, i2) {
-//	    if ((i+1) % NINTERRUPT == 0) R_CheckUserInterrupt();
-	    x1 = REAL(s1)[i1];
-	    x2 = REAL(s2)[i2];
-	    if (ISNAN(x1) || ISNAN(x2))
-		LOGICAL(ans)[i] = NA_LOGICAL;
-	    else
-		LOGICAL(ans)[i] = (x1 > x2);
-	}
-	break;
-    case LEOP:
-	mod_iterate(n1, n2, i1, i2) {
-//	    if ((i+1) % NINTERRUPT == 0) R_CheckUserInterrupt();
-	    x1 = REAL(s1)[i1];
-	    x2 = REAL(s2)[i2];
-	    if (ISNAN(x1) || ISNAN(x2))
-		LOGICAL(ans)[i] = NA_LOGICAL;
-	    else
-		LOGICAL(ans)[i] = (x1 <= x2);
-	}
-	break;
-    case GEOP:
-	mod_iterate(n1, n2, i1, i2) {
-//	    if ((i+1) % NINTERRUPT == 0) R_CheckUserInterrupt();
-	    x1 = REAL(s1)[i1];
-	    x2 = REAL(s2)[i2];
-	    if (ISNAN(x1) || ISNAN(x2))
-		LOGICAL(ans)[i] = NA_LOGICAL;
-	    else
-		LOGICAL(ans)[i] = (x1 >= x2);
-	}
-	break;
-    }
     UNPROTECT(2);
     return ans;
 }
@@ -516,30 +368,34 @@ static SEXP complex_relop(RELOP_TYPE code, SEXP s1, SEXP s2, SEXP call)
     PROTECT(s2);
     ans = allocVector(LGLSXP, n);
 
+    const Rcomplex *px1 = COMPLEX_RO(s1);
+    const Rcomplex *px2 = COMPLEX_RO(s2);
+    int *pa = LOGICAL(ans);
+
     switch (code) {
     case EQOP:
-	mod_iterate(n1, n2, i1, i2) {
+	MOD_ITERATE2(n, n1, n2, i, i1, i2, {
 //	    if ((i+1) % NINTERRUPT == 0) R_CheckUserInterrupt();
-	    x1 = COMPLEX(s1)[i1];
-	    x2 = COMPLEX(s2)[i2];
+	    x1 = px1[i1];
+	    x2 = px2[i2];
 	    if (ISNAN(x1.r) || ISNAN(x1.i) ||
 		ISNAN(x2.r) || ISNAN(x2.i))
-		LOGICAL(ans)[i] = NA_LOGICAL;
+		pa[i] = NA_LOGICAL;
 	    else
-		LOGICAL(ans)[i] = (x1.r == x2.r && x1.i == x2.i);
-	}
+		pa[i] = (x1.r == x2.r && x1.i == x2.i);
+	});
 	break;
     case NEOP:
-	mod_iterate(n1, n2, i1, i2) {
+	MOD_ITERATE2(n, n1, n2, i, i1, i2, {
 //	    if ((i+1) % NINTERRUPT == 0) R_CheckUserInterrupt();
-	    x1 = COMPLEX(s1)[i1];
-	    x2 = COMPLEX(s2)[i2];
+	    x1 = px1[i1];
+	    x2 = px2[i2];
 	    if (ISNAN(x1.r) || ISNAN(x1.i) ||
 		ISNAN(x2.r) || ISNAN(x2.i))
-		LOGICAL(ans)[i] = NA_LOGICAL;
+		pa[i] = NA_LOGICAL;
 	    else
-		LOGICAL(ans)[i] = (x1.r != x2.r || x1.i != x2.i);
-	}
+		pa[i] = (x1.r != x2.r || x1.i != x2.i);
+	});
 	break;
     default:
 	/* never happens (-Wall) */
@@ -554,7 +410,7 @@ static SEXP complex_relop(RELOP_TYPE code, SEXP s1, SEXP s2, SEXP call)
    outside the collation domain. */
 static SEXP string_relop(RELOP_TYPE code, SEXP s1, SEXP s2)
 {
-    R_xlen_t i, n, n1, n2, res;
+    R_xlen_t i, n, n1, n2, res, i1, i2;
     SEXP ans, c1, c2;
     const void *vmax = vmaxget(); // for Scollate
 
@@ -564,105 +420,106 @@ static SEXP string_relop(RELOP_TYPE code, SEXP s1, SEXP s2)
     PROTECT(s1);
     PROTECT(s2);
     PROTECT(ans = allocVector(LGLSXP, n));
+    int *pa = LOGICAL(ans);
 
     switch (code) {
     case EQOP:
-	for (i = 0; i < n; i++) {
+	MOD_ITERATE2(n, n1, n2, i, i1, i2, {
 //	    if ((i+1) % NINTERRUPT == 0) R_CheckUserInterrupt();
-	    c1 = STRING_ELT(s1, i % n1);
-	    c2 = STRING_ELT(s2, i % n2);
+	    c1 = STRING_ELT(s1, i1);
+	    c2 = STRING_ELT(s2, i2);
 	    if (c1 == NA_STRING || c2 == NA_STRING)
-		LOGICAL(ans)[i] = NA_LOGICAL;
+		pa[i] = NA_LOGICAL;
 	    else
-		LOGICAL(ans)[i] = Seql(c1, c2) ? 1 : 0;
-	}
+		pa[i] = Seql(c1, c2) ? 1 : 0;
+	});
 	break;
     case NEOP:
-	for (i = 0; i < n; i++) {
+	MOD_ITERATE2(n, n1, n2, i, i1, i2, {
 //	    if ((i+1) % NINTERRUPT == 0) R_CheckUserInterrupt();
-	    c1 = STRING_ELT(s1, i % n1);
-	    c2 = STRING_ELT(s2, i % n2);
+	    c1 = STRING_ELT(s1, i1);
+	    c2 = STRING_ELT(s2, i2);
 	    if (c1 == NA_STRING || c2 == NA_STRING)
-		LOGICAL(ans)[i] = NA_LOGICAL;
+		pa[i] = NA_LOGICAL;
 	    else
-		LOGICAL(ans)[i] = Seql(c1, c2) ? 0 : 1;
-	}
+		pa[i] = Seql(c1, c2) ? 0 : 1;
+	});
 	break;
     case LTOP:
-	for (i = 0; i < n; i++) {
+	MOD_ITERATE2(n, n1, n2, i, i1, i2, {
 //	    if ((i+1) % NINTERRUPT == 0) R_CheckUserInterrupt();
-	    c1 = STRING_ELT(s1, i % n1);
-	    c2 = STRING_ELT(s2, i % n2);
+	    c1 = STRING_ELT(s1, i1);
+	    c2 = STRING_ELT(s2, i2);
 	    if (c1 == NA_STRING || c2 == NA_STRING)
-		LOGICAL(ans)[i] = NA_LOGICAL;
+		pa[i] = NA_LOGICAL;
 	    else if (c1 == c2)
-		LOGICAL(ans)[i] = 0;
+		pa[i] = 0;
 	    else {
 		errno = 0;
 		res = Scollate(c1, c2);
 		if(errno)
-		    LOGICAL(ans)[i] = NA_LOGICAL;
+		    pa[i] = NA_LOGICAL;
 		else
-		    LOGICAL(ans)[i] = (res < 0) ? 1 : 0;
+		    pa[i] = (res < 0) ? 1 : 0;
 	    }
-	}
+	});
 	break;
     case GTOP:
-	for (i = 0; i < n; i++) {
+	MOD_ITERATE2(n, n1, n2, i ,i1, i2, {
 //	    if ((i+1) % NINTERRUPT == 0) R_CheckUserInterrupt();
-	    c1 = STRING_ELT(s1, i % n1);
-	    c2 = STRING_ELT(s2, i % n2);
+	    c1 = STRING_ELT(s1, i1);
+	    c2 = STRING_ELT(s2, i2);
 	    if (c1 == NA_STRING || c2 == NA_STRING)
-		LOGICAL(ans)[i] = NA_LOGICAL;
+		pa[i] = NA_LOGICAL;
 	    else if (c1 == c2)
-		LOGICAL(ans)[i] = 0;
+		pa[i] = 0;
 	    else {
 		errno = 0;
 		res = Scollate(c1, c2);
 		if(errno)
-		    LOGICAL(ans)[i] = NA_LOGICAL;
+		    pa[i] = NA_LOGICAL;
 		else
-		    LOGICAL(ans)[i] = (res > 0) ? 1 : 0;
+		    pa[i] = (res > 0) ? 1 : 0;
 	    }
-	}
+	});
 	break;
     case LEOP:
-	for (i = 0; i < n; i++) {
+	MOD_ITERATE2(n, n1, n2, i, i1, i2, {
 //	    if ((i+1) % NINTERRUPT == 0) R_CheckUserInterrupt();
-	    c1 = STRING_ELT(s1, i % n1);
-	    c2 = STRING_ELT(s2, i % n2);
+	    c1 = STRING_ELT(s1, i1);
+	    c2 = STRING_ELT(s2, i2);
 	    if (c1 == NA_STRING || c2 == NA_STRING)
-		LOGICAL(ans)[i] = NA_LOGICAL;
+		pa[i] = NA_LOGICAL;
 	    else if (c1 == c2)
-		LOGICAL(ans)[i] = 1;
+		pa[i] = 1;
 	    else {
 		errno = 0;
 		res = Scollate(c1, c2);
 		if(errno)
-		    LOGICAL(ans)[i] = NA_LOGICAL;
+		    pa[i] = NA_LOGICAL;
 		else
-		    LOGICAL(ans)[i] = (res <= 0) ? 1 : 0;
+		    pa[i] = (res <= 0) ? 1 : 0;
 	    }
-	}
+	});
 	break;
     case GEOP:
-	for (i = 0; i < n; i++) {
+	MOD_ITERATE2(n, n1, n2, i, i1, i2, {
 //	    if ((i+1) % NINTERRUPT == 0) R_CheckUserInterrupt();
-	    c1 = STRING_ELT(s1, i % n1);
-	    c2 = STRING_ELT(s2, i % n2);
+	    c1 = STRING_ELT(s1, i1);
+	    c2 = STRING_ELT(s2, i2);
 	    if (c1 == NA_STRING || c2 == NA_STRING)
-		LOGICAL(ans)[i] = NA_LOGICAL;
+		pa[i] = NA_LOGICAL;
 	    else if (c1 == c2)
-		LOGICAL(ans)[i] = 1;
+		pa[i] = 1;
 	    else {
 		errno = 0;
 		res = Scollate(c1, c2);
 		if(errno)
-		    LOGICAL(ans)[i] = NA_LOGICAL;
+		    pa[i] = NA_LOGICAL;
 		else
-		    LOGICAL(ans)[i] = (res >= 0) ? 1 : 0;
+		    pa[i] = (res >= 0) ? 1 : 0;
 	    }
-	}
+	});
 	break;
     }
     UNPROTECT(3);
@@ -683,54 +540,58 @@ static SEXP raw_relop(RELOP_TYPE code, SEXP s1, SEXP s2)
     PROTECT(s2);
     ans = allocVector(LGLSXP, n);
 
+    const Rbyte *px1 = RAW_RO(s1);
+    const Rbyte *px2 = RAW_RO(s2);
+    int *pa = LOGICAL(ans);
+
     switch (code) {
     case EQOP:
-	mod_iterate(n1, n2, i1, i2) {
+	MOD_ITERATE2(n, n1, n2, i, i1, i2, {
 //	    if ((i+1) % NINTERRUPT == 0) R_CheckUserInterrupt();
-	    x1 = RAW(s1)[i1];
-	    x2 = RAW(s2)[i2];
-	    LOGICAL(ans)[i] = (x1 == x2);
-	}
+	    x1 = px1[i1];
+	    x2 = px2[i2];
+	    pa[i] = (x1 == x2);
+	});
 	break;
     case NEOP:
-	mod_iterate(n1, n2, i1, i2) {
+	MOD_ITERATE2(n, n1, n2, i, i1, i2, {
 //	    if ((i+1) % NINTERRUPT == 0) R_CheckUserInterrupt();
-	    x1 = RAW(s1)[i1];
-	    x2 = RAW(s2)[i2];
-	    LOGICAL(ans)[i] = (x1 != x2);
-	}
+	    x1 = px1[i1];
+	    x2 = px2[i2];
+	    pa[i] = (x1 != x2);
+	});
 	break;
     case LTOP:
-	mod_iterate(n1, n2, i1, i2) {
+	MOD_ITERATE2(n, n1, n2, i, i1, i2, {
 //	    if ((i+1) % NINTERRUPT == 0) R_CheckUserInterrupt();
-	    x1 = RAW(s1)[i1];
-	    x2 = RAW(s2)[i2];
-	    LOGICAL(ans)[i] = (x1 < x2);
-	}
+	    x1 = px1[i1];
+	    x2 = px2[i2];
+	    pa[i] = (x1 < x2);
+	});
 	break;
     case GTOP:
-	mod_iterate(n1, n2, i1, i2) {
+	MOD_ITERATE2(n, n1, n2, i, i1, i2, {
 //	    if ((i+1) % NINTERRUPT == 0) R_CheckUserInterrupt();
-	    x1 = RAW(s1)[i1];
-	    x2 = RAW(s2)[i2];
-	    LOGICAL(ans)[i] = (x1 > x2);
-	}
+	    x1 = px1[i1];
+	    x2 = px2[i2];
+	    pa[i] = (x1 > x2);
+	});
 	break;
     case LEOP:
-	mod_iterate(n1, n2, i1, i2) {
+	MOD_ITERATE2(n, n1, n2, i, i1, i2, {
 //	    if ((i+1) % NINTERRUPT == 0) R_CheckUserInterrupt();
-	    x1 = RAW(s1)[i1];
-	    x2 = RAW(s2)[i2];
-	    LOGICAL(ans)[i] = (x1 <= x2);
-	}
+	    x1 = px1[i1];
+	    x2 = px2[i2];
+	    pa[i] = (x1 <= x2);
+	});
 	break;
     case GEOP:
-	mod_iterate(n1, n2, i1, i2) {
+	MOD_ITERATE2(n, n1, n2, i, i1, i2, {
 //	    if ((i+1) % NINTERRUPT == 0) R_CheckUserInterrupt();
-	    x1 = RAW(s1)[i1];
-	    x2 = RAW(s2)[i2];
-	    LOGICAL(ans)[i] = (x1 >= x2);
-	}
+	    x1 = px1[i1];
+	    x2 = px2[i2];
+	    pa[i] = (x1 >= x2);
+	});
 	break;
     }
     UNPROTECT(2);
@@ -740,19 +601,25 @@ static SEXP raw_relop(RELOP_TYPE code, SEXP s1, SEXP s2)
 
 static SEXP bitwiseNot(SEXP a)
 {
+    SEXP ans;
     int np = 0;
     if(isReal(a)) {a = PROTECT(coerceVector(a, INTSXP)); np++;}
-    R_xlen_t i, m = XLENGTH(a);
-    SEXP ans = allocVector(TYPEOF(a), m);
+
     switch(TYPEOF(a)) {
     case INTSXP:
-	for(i = 0; i < m; i++) {
-	    int aa = INTEGER(a)[i];
-	    INTEGER(ans)[i] = (aa == NA_INTEGER) ? aa : ~aa;
+	{
+	    R_xlen_t m = XLENGTH(a);
+	    ans = allocVector(INTSXP, m);
+	    int *pans = INTEGER(ans);
+	    const int *pa = INTEGER_RO(a);
+	    for(R_xlen_t i = 0; i < m; i++) {
+		int aa = pa[i];
+		pans[i] = (aa == NA_INTEGER) ? aa : ~aa;
+	    }
 	}
 	break;
     default:
-	UNIMPLEMENTED_TYPE("bitNot", a);
+	UNIMPLEMENTED_TYPE("bitwNot", a);
     }
     if(np) UNPROTECT(np);
     return ans;
@@ -760,24 +627,33 @@ static SEXP bitwiseNot(SEXP a)
 
 #define mymax(x, y) ((x >= y) ? x : y)
 
-#define BIT(op, name) \
-    int np = 0; \
-    if(isReal(a)) {a = PROTECT(coerceVector(a, INTSXP)); np++;} \
-    if(isReal(b)) {b = PROTECT(coerceVector(b, INTSXP)); np++;} \
-    if (TYPEOF(a) != TYPEOF(b)) error(_("'a' and 'b' must have the same type"));  \
-    R_xlen_t i, m = XLENGTH(a), n = XLENGTH(b), mn = (m && n) ? mymax(m, n) : 0;  \
-    SEXP ans = allocVector(TYPEOF(a), mn); \
-    switch(TYPEOF(a)) { \
-    case INTSXP: \
-	for(i = 0; i < mn; i++) { \
-	    int aa = INTEGER(a)[i%m], bb =  INTEGER(b)[i%n]; \
-	    INTEGER(ans)[i] = (aa == NA_INTEGER || bb == NA_INTEGER) ? NA_INTEGER : aa op bb; \
-	} \
-	break; \
-    default: \
-	UNIMPLEMENTED_TYPE(name, a); \
-    } \
-    if(np) UNPROTECT(np); \
+#define BIT(op, name)							\
+    SEXP ans;								\
+    int np = 0;								\
+    if(isReal(a)) {a = PROTECT(coerceVector(a, INTSXP)); np++;}		\
+    if(isReal(b)) {b = PROTECT(coerceVector(b, INTSXP)); np++;}		\
+    if (TYPEOF(a) != TYPEOF(b))						\
+	error(_("'a' and 'b' must have the same type"));		\
+    switch(TYPEOF(a)) {							\
+    case INTSXP:							\
+	{								\
+	    R_xlen_t i, ia, ib;						\
+	    R_xlen_t m = XLENGTH(a), n = XLENGTH(b),			\
+		mn = (m && n) ? mymax(m, n) : 0;			\
+	    ans = allocVector(INTSXP, mn);				\
+	    int *pans = INTEGER(ans);					\
+	    const int *pa = INTEGER_RO(a), *pb = INTEGER_RO(b);		\
+	    MOD_ITERATE2(mn, m, n, i, ia, ib, {				\
+		    int aa = pa[ia]; int bb = pb[ib];			\
+		    pans[i] = (aa == NA_INTEGER || bb == NA_INTEGER) ?	\
+			NA_INTEGER : aa op bb;				\
+		});							\
+	}								\
+	break;								\
+    default:								\
+	UNIMPLEMENTED_TYPE(name, a);					\
+    }									\
+    if(np) UNPROTECT(np);						\
     return ans
 
 static SEXP bitwiseAnd(SEXP a, SEXP b)
@@ -797,18 +673,29 @@ static SEXP bitwiseXor(SEXP a, SEXP b)
 
 static SEXP bitwiseShiftL(SEXP a, SEXP b)
 {
+    SEXP ans;
     int np = 0;
     if(isReal(a)) {a = PROTECT(coerceVector(a, INTSXP)); np++;}
     if(!isInteger(b)) {b = PROTECT(coerceVector(b, INTSXP)); np++;}
-    R_xlen_t i, m = XLENGTH(a), n = XLENGTH(b), 
-	mn = (m && n) ? mymax(m, n) : 0;
-    SEXP ans = allocVector(TYPEOF(a), mn);
+    if (TYPEOF(a) != TYPEOF(b))
+	error(_("'a' and 'b' must have the same type"));
+
     switch(TYPEOF(a)) {
     case INTSXP:
-	for(i = 0; i < mn; i++) {
-	    int aa = INTEGER(a)[i%m], bb = INTEGER(b)[i%n];
-	    INTEGER(ans)[i] = 
-		(aa == NA_INTEGER || bb == NA_INTEGER || bb < 0 || bb > 31) ? NA_INTEGER : ((unsigned int)aa << bb);
+	{
+	    R_xlen_t i, ia, ib;
+	    R_xlen_t m = XLENGTH(a), n = XLENGTH(b),
+		mn = (m && n) ? mymax(m, n) : 0;
+	    ans = allocVector(INTSXP, mn);
+	    int *pans = INTEGER(ans);
+	    const int *pa = INTEGER_RO(a), *pb = INTEGER_RO(b);
+	    MOD_ITERATE2(mn, m, n, i, ia, ib, {
+		    int aa = pa[ia]; int bb = pb[ib];
+		    pans[i] =
+			(aa == NA_INTEGER || bb == NA_INTEGER ||
+			 bb < 0 || bb > 31) ?
+			NA_INTEGER : ((unsigned int)aa << bb);
+		});
 	}
 	break;
     default:
@@ -820,18 +707,29 @@ static SEXP bitwiseShiftL(SEXP a, SEXP b)
 
 static SEXP bitwiseShiftR(SEXP a, SEXP b)
 {
+    SEXP ans;
     int np = 0;
     if(isReal(a)) {a = PROTECT(coerceVector(a, INTSXP)); np++;}
     if(!isInteger(b)) {b = PROTECT(coerceVector(b, INTSXP)); np++;}
-    R_xlen_t i, m = XLENGTH(a), n = XLENGTH(b), 
-	mn = (m && n) ? mymax(m, n) : 0;
-    SEXP ans = allocVector(TYPEOF(a), mn);
+    if (TYPEOF(a) != TYPEOF(b))
+	error(_("'a' and 'b' must have the same type"));
+
     switch(TYPEOF(a)) {
     case INTSXP:
-	for(i = 0; i < mn; i++) {
-	    int aa = INTEGER(a)[i%m], bb = INTEGER(b)[i%n];
-	    INTEGER(ans)[i] = 
-		(aa == NA_INTEGER || bb == NA_INTEGER || bb < 0 || bb > 31) ? NA_INTEGER : ((unsigned int)aa >> bb);
+	{
+	    R_xlen_t i, ia, ib;
+	    R_xlen_t m = XLENGTH(a), n = XLENGTH(b),
+		mn = (m && n) ? mymax(m, n) : 0;
+	    ans = allocVector(TYPEOF(a), mn);
+	    int *pans = INTEGER(ans);
+	    const int *pa = INTEGER_RO(a), *pb = INTEGER_RO(b);
+	    MOD_ITERATE2(mn, m, n, i, ia, ib, {
+		    int aa = pa[ia]; int bb = pb[ib];
+		    pans[i] =
+			(aa == NA_INTEGER || bb == NA_INTEGER ||
+			 bb < 0 || bb > 31) ?
+			NA_INTEGER : ((unsigned int)aa >> bb);
+		});
 	}
 	break;
     default:
